﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.Linq.Expressions;
using System.Collections;

class Person : IComparable<Person> {
	public string FirstName { get; set; }
	public string LastName { get; set; }
	public DateTime Birthday { get; set; }

	public override string ToString() {
		return $"[ {nameof(FirstName)} = {FirstName}, {nameof(LastName)} = {LastName}, {nameof(Birthday)} = {Birthday.ToShortDateString()} ]";
	}

	public int CompareTo(Person other) {
		return LastName.CompareTo(other.LastName);
	}
}

class Program {


	static void Main(string[] args) {
		Console.WriteLine($"64-bit: {Environment.Is64BitProcess}");

		SortedList<string, int> l = new SortedList<string, int> { { "bb", 2 }, { "cc", 3 }, { "aa", 1 } };

		SortedSet<string> set = new SortedSet<string> { "bb", "aa", "cc", "aa" };

		{
			var personSet = LinqAwareSortedSet<Person>.CreateSmart(p => p.LastName);

			var sw = System.Diagnostics.Stopwatch.StartNew();
			GenerateItems(personSet);
			sw.Stop();
			Console.WriteLine($"Initial sorting add took: {sw.Elapsed}");

			sw = System.Diagnostics.Stopwatch.StartNew();
			foreach (var p in personSet.OrderBy(p => p.LastName).Take(10)) {
				Console.WriteLine(p);
			}
			sw.Stop();
			Console.WriteLine($"Enumeration took: {sw.Elapsed}");
		}

		Console.WriteLine();

		{
			var personSet = LinqAwareSortedSet<Person>.Create(p => p.LastName);

			var sw = System.Diagnostics.Stopwatch.StartNew();
			GenerateItems(personSet);
			sw.Stop();
			Console.WriteLine($"Initial sorting add took: {sw.Elapsed}");

			sw = System.Diagnostics.Stopwatch.StartNew();
			foreach (var p in personSet.OrderBy(p => p.LastName).Take(10)) {
				Console.WriteLine(p);
			}
			sw.Stop();
			Console.WriteLine($"Enumeration took: {sw.Elapsed}");
		}

		Console.WriteLine();

		{
			var personSet = LinqAwareSortedSet<Person>.CreateDefault();

			var sw = System.Diagnostics.Stopwatch.StartNew();
			GenerateItems(personSet);
			sw.Stop();
			Console.WriteLine($"Initial sorting add took: {sw.Elapsed}");

			sw = System.Diagnostics.Stopwatch.StartNew();
			foreach (var p in personSet.OrderBy(p => p.LastName).Take(10)) {
				Console.WriteLine(p);
			}
			sw.Stop();
			Console.WriteLine($"Enumeration took: {sw.Elapsed}");
		}

	}

	static void GenerateItems(LinqAwareSortedSet<Person> set) {
		for (int i = 1; i <= 1000000; i++) {
			var sb = new StringBuilder();
			int x = i;
			do {
				sb.Append((char) ('A' + x % 26));
				x /= 26;
			} while (x > 1);
			set.Add(new Person { FirstName = sb.ToString(), LastName = sb.ToString().ToLower(), Birthday = DateTime.Today.AddDays(i) });
		}
	}
}

class LinqAwareSortedSet<T> : IEnumerable<T> {
	private SortedSet<T> data;
	private System.Reflection.MemberInfo keyMemberInfo;

	private LinqAwareSortedSet() {
	}

	public static LinqAwareSortedSet<T> Create<TKey>(Func<T, TKey> keySelector) {
		var linqSet = new LinqAwareSortedSet<T>();
		linqSet.data = new SortedSet<T>(
			Comparer<T>.Create((v1, v2) => Comparer<TKey>.Default.Compare(keySelector(v1), keySelector(v2)))
		);
		return linqSet;
	}

	public static LinqAwareSortedSet<T> CreateSmart<TKey>(Expression<Func<T, TKey>> keySelectorExpr) {
		var linqSet = Create(keySelectorExpr.Compile());
		linqSet.keyMemberInfo = ((keySelectorExpr as LambdaExpression)?.Body as MemberExpression)?.Member;
		return linqSet;
	}

	public static LinqAwareSortedSet<T> CreateDefault() {
		var linqSet = new LinqAwareSortedSet<T>();
		linqSet.data = new SortedSet<T>();
		return linqSet;
	}

	public void Add(T item) {
		data.Add(item);
	}

	public IEnumerable<T> OrderBy<TKey>(Expression<Func<T, TKey>> keySelectorExpr) {
		if (keyMemberInfo == ((keySelectorExpr as LambdaExpression)?.Body as MemberExpression)?.Member) {
			Console.WriteLine("LinqAwareSortedSet.OrderBy: optimized query without re-sorting");
			return data;
		} else {
			Console.WriteLine("LinqAwareSortedSet.OrderBy: fallback to re-sorting query");
			return data.OrderBy(keySelectorExpr.Compile());
		}
	}

	public IEnumerator<T> GetEnumerator() {
		return ((IEnumerable<T>) data).GetEnumerator();
	}

	IEnumerator IEnumerable.GetEnumerator() {
		return ((IEnumerable<T>) data).GetEnumerator();
	}
}